# validated: 2017-10-08 DS 8a37b81f4e10 cpp/Storage.cpp cpp/Storage.h cpp/Storage_load.cpp cpp/Storage_save.cpp
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import contextlib
import os
import threading

from .message import Message
from .network_connection import NetworkConnection
from .storage_load import load_entries
from .storage_save import save_entries
from .structs import EntryInfo, ConnectionInfo
from .value import Value

from .support.compat import monotonic, file_replace
from .support.lists import ensure_id_exists

from .constants import (
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    kExecuteRpc,
    kRpcResponse,
    
    NT_UNASSIGNED,
    NT_PERSISTENT,
    
    NT_RPC,
    
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS
)

import logging
logger = logging.getLogger('nt')



class _Entry(object):
    __slots__ = ['name', '_value', 'flags',
                 'id', 'local_id', 'seq_num', 'local_write',
                 'rpc_uid', 'rpc_call_uid', 'user_entry']
    
    def __init__(self, name, local_id, user_entry):
        
        # We redundantly store the name so that it's available when accessing the
        # raw Entry* via the ID map.
        self.name = name
        
        # The current value and flags.
        self._value = None
        self.flags = 0
        
        # Unique ID for self entry as used in network messages.  The value is
        # assigned by the server, on the client this is 0xffff until an
        # entry assignment is received back from the server.
        self.id = 0xffff
        
        # Local ID
        self.local_id = local_id
        
        # Sequence number for update resolution.
        self.seq_num = 0
        
        # If value has been written locally.  Used during initial handshake
        # on client to determine whether or not to accept remote changes.
        self.local_write = False
        
        # RPC handle
        self.rpc_uid = None
        
        # Last UID used when calling self RPC (primarily for client use).  This
        # is incremented for each call.
        self.rpc_call_uid = 0
        
        # python-specific: User-visible entry for optimized value retrieval
        self.user_entry = user_entry
        
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        self._value = value
        self.user_entry._value = value
        
    def isPersistent(self):
        return (self.flags & NT_PERSISTENT) != 0 
    
    def increment_seqnum(self):
        self.seq_num += 1
        self.seq_num &= 0xffff
        
    def isSeqNewerThan(self, other):
        '''
            self > other
        '''
        seq_num = self.seq_num
        if seq_num < other:
            return (other - seq_num) > 32768
        elif seq_num > other:
            return (seq_num - other) < 32768
        else:
            return False
        
    def isSeqNewerOrEqual(self, other):
        '''
            self >= other
        '''
        seq_num = self.seq_num
        if seq_num < other:
            return (other - seq_num) > 32768
        elif seq_num > other:
            return (seq_num - other) < 32768
        else:
            return True
        
    def isRpc(self):
        return self.value.type == NT_RPC
        
    def __repr__(self):
        return "<_Entry name='%s' value=%s flags=%s id=%s local_id=%s seq_num=%s local_write=%s rpc_uid=%s rpc_call_uid=%s" % \
            (self.name, self.value, self.flags, self.id, self.local_id,
             self.seq_num, self.local_write, self.rpc_uid, self.rpc_call_uid)


class Storage(object):
    
    def __init__(self, entry_notifier, rpc_server, user_entry_creator):
        self.m_notifier = entry_notifier
        self.m_rpc_server = rpc_server
        
        # python-specific
        self.m_user_entry_creator = user_entry_creator
        
        self.m_mutex = threading.Lock()
        self.m_entries = {}
        self.m_idmap = []
        self.m_localmap = []
        self.m_rpc_results = {}
        self.m_rpc_blocking_calls = set()
        
        # If any persistent values have changed
        self.m_persistent_dirty = False
        
        # condition variable and termination flag for blocking on a RPC result
        self.m_terminating = False
        self.m_rpc_results_cond = threading.Condition(self.m_mutex)
        
        # configured by dispatcher at startup
        self.m_dispatcher = None
        self.m_server = True
        
        # Differs from ntcore because python doesn't have switch statements...
        self._process_fns = {
            kEntryAssign:   self._processIncomingEntryAssign,
            kEntryUpdate:   self._processIncomingEntryUpdate,
            kFlagsUpdate:   self._processIncomingFlagsUpdate,
            kEntryDelete:   self._processIncomingEntryDelete,
            kClearEntries:  self._processIncomingClearEntries,
            kExecuteRpc:    self._processIncomingExecuteRpc,
            kRpcResponse:   self._processIncomingRpcResponse
        }
    
    def stop(self):
        self.m_terminating = True
        with self.m_mutex:
            self.m_rpc_results_cond.notify_all()
    
    def setDispatcher(self, dispatcher, server):
        with self.m_mutex:
            self.m_dispatcher = dispatcher
            self.m_server = server
    
    def clearDispatcher(self):
        self.m_dispatcher = None
    
    def getMessageEntryType(self, msg_id):
        with self.m_mutex:
            if msg_id >= len(self.m_idmap):
                return NT_UNASSIGNED
        
            entry = self.m_idmap[msg_id]
            if not entry or not entry.value:
                return NT_UNASSIGNED
        
            return entry.value.type
    
    @contextlib.contextmanager
    def _lockAndGetSendQueue(self):
        '''
            Python specific function
            
            This exists because our lock object isn't like the C++ lock
        '''
        with self.m_mutex:
            dispatcher = self.m_dispatcher
            if not dispatcher:
                yield None
                return
            
            queue_outgoing = dispatcher._queueOutgoing
            outgoing = []
            yield outgoing
        
        # This has to happen outside the lock
        for o in outgoing:
            try:
                queue_outgoing(*o)
            except TypeError:
                raise
    
    def processIncoming(self, msg, conn):
        # Note: c++ version takes a third param (weak_conn), but it's
        #       not needed here as conn == weak_conn
        fn = self._process_fns.get(msg.type)
        if fn:
            with self._lockAndGetSendQueue() as outgoing:
                fn(msg, conn, outgoing)
    
    def _processIncomingEntryAssign(self, msg, conn, outgoing):

        msg_id = msg.id
        name = msg.str
        entry = None
        may_need_update = False
        
        if self.m_server:
            # if we're a server, id=0xffff requests are requests for an id
            # to be assigned, we need to send the assignment back to
            # the sender as well as all other connections.
            if msg_id == 0xffff:
                entry = self._getOrNew(name)
                # see if it was already assigned; ignore if so.
                if entry.id != 0xffff:
                    return
                
                entry.flags = msg.flags
                entry.seq_num = msg.seq_num_uid
                self._setEntryValueImpl(entry, msg.value, outgoing, False)
                return
            
            if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
                # ignore arbitrary entry assignments
                # this can happen due to e.g. assignment to deleted entry
                logger.debug("server: received assignment to unknown entry")
                return

            entry = self.m_idmap[msg_id]
        else:
            # clients simply accept assignments
            if msg_id == 0xffff:
                logger.debug("client: received entry assignment request?")
                return

            ensure_id_exists(self.m_idmap, msg_id)
            
            entry = self.m_idmap[msg_id]
            if entry is None:
                # create local
                entry = self._getOrNew(name)
                entry.id = msg_id
                self.m_idmap[msg_id] = entry
                if entry.value is None:
                    # didn't exist at all (rather than just being a response to a
                    # id assignment request)
                    entry.value = msg.value
                    entry.flags = msg.flags
                    entry.seq_num = msg.seq_num_uid
                    
                    # notify
                    self.m_notifier.notifyEntry(entry.local_id, name, entry.value, NT_NOTIFY_NEW)
                    return

                may_need_update = True  # we may need to send an update message
                
                # if the received flags don't match what we sent, most likely
                # updated flags locally in the interim; send flags update message.
                if msg.flags != entry.flags:
                    outmsg = Message.flagsUpdate(msg_id, entry.flags)
                    outgoing.append((outmsg, None, None))
        
        # common client and server handling

        # already exists; ignore if sequence number not higher than local
        seq_num = msg.seq_num_uid
        if entry.isSeqNewerThan(seq_num):
            if may_need_update:
                outmsg = Message.entryUpdate(entry.id, entry.seq_num,
                                             entry.value)
                outgoing.append((outmsg, None, None))
            
            return

        # sanity check: name should match id
        if msg.str != entry.name:
            logger.debug("entry assignment for same id with different name?")
            return

        notify_flags = NT_NOTIFY_UPDATE

        # don't update flags from a <3.0 remote (not part of message)
        # don't update flags if self is a server response to a client id request
        if not may_need_update and conn.get_proto_rev() >= 0x0300:
            # update persistent dirty flag if persistent flag changed
            if (entry.flags & NT_PERSISTENT) != (msg.flags & NT_PERSISTENT):
                self.m_persistent_dirty = True

            if entry.flags != msg.flags:
                notify_flags |= NT_NOTIFY_FLAGS

            entry.flags = msg.flags

        # update persistent dirty flag if the value changed and it's persistent
        if entry.isPersistent() and entry.value != msg.value:
            self.m_persistent_dirty = True

        # update local
        entry.value = msg.value
        entry.seq_num = seq_num

        # notify
        self.m_notifier.notifyEntry(entry.local_id, name, entry.value, notify_flags)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outmsg = Message.entryAssign(entry.name, msg_id, msg.seq_num_uid,
                                         msg.value, entry.flags)
            outgoing.append((outmsg, None, conn))
    
    def _processIncomingEntryUpdate(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates;
            # this can happen due to deleted entries
            logger.debug("received update to unknown entry")
            return

        entry = self.m_idmap[msg_id]

        # ignore if sequence number not higher than local
        seq_num = msg.seq_num_uid
        if entry.isSeqNewerOrEqual(seq_num):
            return
        
        # update local
        entry.value = msg.value
        entry.seq_num = seq_num

        # update persistent dirty flag if it's a persistent value
        if entry.isPersistent():
            self.m_persistent_dirty = True
        
        # notify
        self.m_notifier.notifyEntry(entry.local_id, entry.name, entry.value,
                                    NT_NOTIFY_UPDATE)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processIncomingFlagsUpdate(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates
            # this can happen due to deleted entries
            logger.debug("received flags update to unknown entry")
            return

        self._setEntryFlagsImpl(self.m_idmap[msg_id], msg.flags, outgoing, False)
        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processIncomingEntryDelete(self, msg, conn, outgoing):
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore arbitrary entry updates
            # this can happen due to deleted entries
            logger.debug("received delete to unknown entry")
            return
        
        self._deleteEntryImpl(self.m_idmap[msg_id], outgoing, False)
        
        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processIncomingClearEntries(self, msg, conn, outgoing):
        # update local
        self._deleteAllEntriesImpl(False)

        # broadcast to all other connections (note for client there won't
        # be any other connections, don't bother)
        if self.m_server and outgoing is not None:
            outgoing.append((msg, None, conn))
    
    def _processIncomingExecuteRpc(self, msg, conn, outgoing):
        if not self.m_server:
            return    # only process on server

        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore call to non-existent RPC
            # this can happen due to deleted entries
            logger.debug("received RPC call to unknown entry")
            return

        entry = self.m_idmap[msg_id]
        if not entry.value or not entry.isRpc():
            logger.debug("received RPC call to non-RPC entry")
            return

        call_uid = msg.seq_num_uid
        # XXX: TODO
        assert False
        self.m_rpc_server.processRpc(entry.local_id, entry.rpc_call_uid,
                                     entry.name, msg, entry.rpc_callback,
                                     conn.uid(), conn.queueOutgoing, conn.info())
    
    def _processIncomingRpcResponse(self, msg, conn, outgoing):
        if self.m_server:
            return    # only process on client
        
        msg_id = msg.id
        if msg_id >= len(self.m_idmap) or not self.m_idmap[msg_id]:
            # ignore call to non-existent RPC
            # this can happen due to deleted entries
            logger.debug("received RPC response to unknown entry")
            return
        
        entry = self.m_idmap[msg_id]
        if not entry.value or not entry.isRpc():
            logger.debug("received RPC call to non-RPC entry")
            return
        
        self.m_rpc_results[(entry.local_id, msg.seq_num_uid)] = msg.str
        self.m_rpc_results_cond.notify_all()
    
    def getInitialAssignments(self, conn, msgs):
        with self.m_mutex:
            conn.set_state(NetworkConnection.State.kSynchronized)
            for entry in self.m_entries.values():
                if entry.value is None:
                    continue
                msgs.append(Message.entryAssign(entry.name, entry.id,
                                                entry.seq_num,
                                                entry.value, entry.flags))
    
    def applyInitialAssignments(self, conn, msgs, new_server, out_msgs):
        with self._lockAndGetSendQueue() as update_msgs:
            if self.m_server:
                return    # should not do this on server
        
            conn.set_state(NetworkConnection.State.kSynchronized)
        
            # clear existing id's
            for entry in self.m_entries.values():
                entry.id = 0xffff
        
            # clear existing idmap
            del self.m_idmap[:]
        
            # apply assignments
            for msg in msgs:
                if msg.type != kEntryAssign:
                    logger.debug("client: received non-entry assignment request?")
                    continue
        
                msg_id = msg.id
                if msg_id == 0xffff:
                    logger.debug("client: received entry assignment request?")
                    continue
        
                seq_num = msg.seq_num_uid
                name = msg.str
        
                entry = self._getOrNew(name)
                if entry.value is None:
                    entry.value = msg.value
                    entry.flags = msg.flags
                    entry.seq_num = msg.seq_num_uid
                
                    # notify
                    self.m_notifier.notifyEntry(entry.local_id, name, entry.value, NT_NOTIFY_NEW)
                else:
                    # if we have written the value locally and the value is not persistent,
                    # then we don't update the local value and instead send it back to the
                    # server as an update message
                    if entry.local_write and not entry.isPersistent():
                        update_msgs.append((Message.entryUpdate(entry.id, entry.seq_num, entry.value), None, None))
                    else:
                        entry.value = msg.value
                        entry.seq_num = seq_num
                        notify_flags = NT_NOTIFY_UPDATE
                        # don't update flags from a <3.0 remote (not part of message)
                        if conn.get_proto_rev() >= 0x0300:
                            if entry.flags != msg.flags:
                                notify_flags |= NT_NOTIFY_FLAGS
        
                            entry.flags = msg.flags
        
                        # notify
                        self.m_notifier.notifyEntry(entry.local_id, name, entry.value, notify_flags)
        
                # set id and save to idmap
                entry.id = msg_id
                
                ensure_id_exists(self.m_idmap, msg_id)
                self.m_idmap[msg_id] = entry
            
            # delete or generate assign messages for unassigned local entries
            def _shouldDelete(entry):
                #  was assigned by the server, don't delete
                if entry.id != 0xffff:
                    return False
                
                # if we have written the value locally, we send an assign message to the
                # server instead of deleting
                if entry.local_write:
                    out_msgs.append(Message.entryAssign(entry.name, entry.id,
                                                        entry.seq_num,
                                                        entry.value, entry.flags))
                    return False
                  
                # otherwise delete
                return True
            
            self._deleteAllEntriesImpl(False, _shouldDelete)
    
    def getEntryValue(self, name):
        with self.m_mutex:
            e = self.m_entries.get(name)
            if e:
                return e.value
    
    def setDefaultEntryValue(self, name, value):
        if not name:
            return False    # can't compare empty name
        
        if value is None:
            return False    # can't compare to a null value
    
        with self._lockAndGetSendQueue() as outgoing:
            entry = self._getOrNew(name)
            
            # We return early if value already exists; if types match return true
            if entry.value is not None:
                return entry.value.type == value.type
            
            self._setEntryValueImpl(entry, value, outgoing, True)
            return True
    
    def setDefaultEntryValueById(self, local_id, value):
        if value is None:
            return False    # can't compare to a null value
    
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return False
            
            # We return early if value already exists; if types match return true
            if entry.value:
                return entry.value.type == value.type
            
            self._setEntryValueImpl(entry, value, outgoing, True)
            return True
    
    def setEntryValue(self, name, value):
        if not name:
            return True
        if value is None:
            return True
        
        with self._lockAndGetSendQueue() as outgoing:
            entry = self._getOrNew(name)
            if entry.value is not None and entry.value.type != value.type:
                return False # error on type mismatch
            
            self._setEntryValueImpl(entry, value, outgoing, True)
            return True
    
    def setEntryValueById(self, local_id, value):
        if value is None:
            return True
        
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return True
            
            if entry.value and entry.value.type != value.type:
                return False # error on type mismatch
            
            self._setEntryValueImpl(entry, value, outgoing, True)
            return True
    
    def _setEntryValueImpl(self, entry, value, outgoing, local):
        
        if value is None:
            return True
        
        old_value = entry.value
        entry.value = value
    
        # if we're the server, assign an id if it doesn't have one
        if self.m_server and entry.id == 0xffff:
            entry.id = len(self.m_idmap)
            self.m_idmap.append(entry)
        
        # update persistent dirty flag if value changed and it's persistent
        if entry.isPersistent() and (old_value is None or old_value != value):
            self.m_persistent_dirty = True
        
        # notify
        nflag = NT_NOTIFY_LOCAL if local else 0
        if old_value is None:
            self.m_notifier.notifyEntry(entry.local_id, entry.name, value, NT_NOTIFY_NEW | nflag)

        elif old_value != value:
            self.m_notifier.notifyEntry(entry.local_id, entry.name, value, NT_NOTIFY_UPDATE | nflag)
        
        # remember local changes
        if local:
            entry.local_write = True
        
        # generate message
        if outgoing is None or (not local and not self.m_server):
            return
        
        if old_value is None or old_value.type != value.type:
            if local:
                entry.increment_seqnum()
            msg = Message.entryAssign(entry.name, entry.id, entry.seq_num,
                                      value, entry.flags)
            outgoing.append((msg, None, None))

        elif old_value != value:
            if local:
                entry.increment_seqnum()
            
            # don't send an update if we don't have an assigned id yet
            if entry.id != 0xffff:
                msg = Message.entryUpdate(entry.id, entry.seq_num, value)
                outgoing.append((msg, None, None))

    def setEntryTypeValue(self, name, value):
        if not name:
            return
        if value is None:
            return
    
        with self._lockAndGetSendQueue() as outgoing:
            entry = self._getOrNew(name)
            self._setEntryValueImpl(entry, value, outgoing, True)
        
    def setEntryTypeValueById(self, local_id, value):
        if value is None:
            return
    
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return
            self._setEntryValueImpl(entry, value, outgoing, True)
    
    def setEntryFlags(self, name, flags):
        if not name:
            return
        
        with self._lockAndGetSendQueue() as outgoing:
            entry = self.m_entries.get(name)
            if entry is None:
                return
            self._setEntryFlagsImpl(entry, flags, outgoing, True)
    
    def setEntryFlagsById(self, local_id, flags):
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return
            self._setEntryFlagsImpl(entry, flags, outgoing, True)
    
    def _setEntryFlagsImpl(self, entry, flags, outgoing, local):
        if entry.value is None or entry.flags == flags:
            return

        # update persistent dirty flag if persistent flag changed
        if (entry.flags & NT_PERSISTENT) != (flags & NT_PERSISTENT):
            self.m_persistent_dirty = True
        
        entry.flags = flags
    
        # notify
        self.m_notifier.notifyEntry(entry.local_id, entry.name, entry.value,
                                    NT_NOTIFY_FLAGS | (NT_NOTIFY_LOCAL if local else 0))
    
        # generate message
        if not local or outgoing is None:
            return
        
        entry_id = entry.id
        
        # don't send an update if we don't have an assigned id yet
        if entry_id != 0xffff:
            outgoing.append((Message.flagsUpdate(entry_id, flags), None, None))
    
    def getEntryFlags(self, name):
        with self.m_mutex:
            entry = self.m_entries.get(name)
            return entry.flags if entry else 0
    
    def getEntryFlagsById(self, local_id):
        with self.m_mutex:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return 0
            else:
                return entry.flags
    
    def deleteEntry(self, name):
        if not name:
            return
        
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_entries[name]
            except KeyError:
                return
            else:
                self._deleteEntryImpl(entry, outgoing, True)
    
    def deleteEntryById(self, local_id):
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return
            else:
                self._deleteEntryImpl(entry, outgoing, True)
            
    def _deleteEntryImpl(self, entry, outgoing, local):
        
        entry_id = entry.id
        if entry_id < len(self.m_idmap):
            self.m_idmap[entry_id] = None
        
        # empty the value and reset id and local_write flag
        old_value = entry.value
        entry.value = None
        entry.id = 0xffff
        entry.local_write = False
            
        # Remove RPC if there was one
        if entry.rpc_uid is not None:
            self.m_rpc_server.removeRpc(entry.rpc_uid)
            entry.rpc_uid = None
        
        # update persistent dirty flag if it's a persistent value
        if entry.isPersistent():
            self.m_persistent_dirty = True
        
        # reset flags
        entry.flags = 0
        
        if old_value is None:
            return  # was not previously assigned
        
        # notify
        self.m_notifier.notifyEntry(entry.local_id, entry.name, old_value,
                                    NT_NOTIFY_DELETE | (NT_NOTIFY_LOCAL if local else 0))
    
        # if it had a value, generate message
        # don't send an update if we don't have an assigned id yet
        if outgoing is not None and local and entry_id != 0xffff:
            outgoing.append((Message.entryDelete(entry_id), None, None))
    
    def _defaultShouldDelete(self, entry):
        return not entry.isPersistent()
    
    def _deleteAllEntriesImpl(self, local, should_delete=None):
        if should_delete is None:
            should_delete = self._defaultShouldDelete
        
        notify_flags = NT_NOTIFY_DELETE | (NT_NOTIFY_LOCAL if local else 0)
        deleted = False
        
        for name, entry in self.m_entries.items():
            if entry.value is not None and should_delete(entry):
                # notify it's being deleted
                self.m_notifier.notifyEntry(entry.local_id, entry.name,
                                            entry.value, notify_flags)
    
                # remove it from idmap
                if entry.id < len(self.m_idmap):
                    self.m_idmap[entry.id] = None
                
                entry.id = 0xffff
                entry.local_write = False
                entry.value = None
                
                deleted = True
        
        return deleted
    
    def deleteAllEntries(self):
        with self._lockAndGetSendQueue() as outgoing:
            deleted = self._deleteAllEntriesImpl(True)
        
            # generate message
            if deleted and outgoing is not None:
                outgoing.append((Message.clearEntries(), None, None))
    
    def _getOrNew(self, name):
        entry = self.m_entries.get(name)
        if not entry:
            local_id = len(self.m_localmap)
            user_entry = self.m_user_entry_creator(name, local_id)
            entry = _Entry(name, local_id, user_entry)
            self.m_entries[name] = entry
            self.m_localmap.append(entry)
        return entry
    
    # ntcore: getEntry
    def getEntryId(self, name):
        if name:
            with self.m_mutex:
                entry = self._getOrNew(name)
                return entry.local_id
    
    # python-specific
    def getEntry(self, name):
        if name:
            with self.m_mutex:
                entry = self._getOrNew(name)
                return entry.user_entry
    
    # python-specific: returns user entries instead of ids
    def getEntries(self, prefix, types):
        with self.m_mutex:
            entries = []
            types = types if isinstance(types, int) else ord(types)
            for k, entry in self.m_entries.items():
                if entry.value is None or not k.startswith(prefix):
                    continue
                if types != 0 and ((types & ord(entry.value.type)) == 0):
                    continue
                entries.append(entry.user_entry)
        return entries
    
    def getEntryInfoById(self, local_id):
        with self.m_mutex:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return EntryInfo(None, NT_UNASSIGNED, 0)
            else:
                return EntryInfo(entry.name, entry.value.type, entry.flags)
    
    def getEntryNameById(self, local_id):
        with self.m_mutex:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return None
            else:
                return entry.name
    
    def getEntryTypeById(self, local_id):
        with self.m_mutex:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return NT_UNASSIGNED
            else:
                if entry.value is None:
                    return NT_UNASSIGNED
                return entry.value.type
    
    def getEntryInfo(self, prefix, types):
        with self.m_mutex:
            infos = []
            types = types if isinstance(types, int) else ord(types)
            for k, entry in self.m_entries.items():
                value = entry.value
                if value is None or not k.startswith(prefix):
                    continue
                
                if types != 0 and (types & ord(value.type)) == 0:
                    continue
        
                info = EntryInfo(entry.name, value.type, entry.flags)
                infos.append(info)
        
            return infos
    
    def addListener(self, prefix, callback, flags):
        with self.m_mutex:
            uid = self.m_notifier.add(callback, prefix, flags)
            # perform immediate notifications
            if (flags & NT_NOTIFY_IMMEDIATE) != 0 and (flags & NT_NOTIFY_NEW) != 0:
                for k, entry in self.m_entries.items():
                    if entry.value is None or not k.startswith(prefix):
                        continue
                    self.m_notifier.notifyEntry(entry.local_id, k, entry.value,
                                                NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW, uid)
        return uid
    
    def addListenerById(self, local_id, callback, flags):
        with self.m_mutex:
            uid = self.m_notifier.addById(callback, local_id, flags)
            # perform immediate notifications
            if (flags & NT_NOTIFY_IMMEDIATE) != 0 and (flags & NT_NOTIFY_NEW) != 0:
                try:
                    entry = self.m_localmap[local_id]
                except IndexError:
                    pass
                else:
                    # if no value, don't notify
                    if entry.value is not None:
                        self.m_notifier.notifyEntry(local_id, entry.name, entry.value,
                                                    NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW, uid)
        return uid
    
    def addPolledListener(self, poller_uid, prefix, flags):
        with self.m_mutex:
            uid = self.m_notifier.addPolled(poller_uid, prefix, flags)
            # perform immediate notifications
            if (flags & NT_NOTIFY_IMMEDIATE) != 0 and (flags & NT_NOTIFY_NEW) != 0:
                for k, entry in self.m_entries.items():
                    if entry.value is None or not k.startswith(prefix):
                        continue
                    self.m_notifier.notifyEntry(entry.local_id, k, entry.value,
                                                NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW, uid)
        return uid
    
    def addPolledListenerById(self, poller_uid, local_id, flags):
        with self.m_mutex:
            uid = self.m_notifier.addPolledById(poller_uid, local_id, flags)
            # perform immediate notifications
            if (flags & NT_NOTIFY_IMMEDIATE) != 0 and (flags & NT_NOTIFY_NEW) != 0:
                try:
                    entry = self.m_localmap[local_id]
                except IndexError:
                    pass
                else:
                    # if no value, don't notify
                    if entry.value is not None:
                        self.m_notifier.notifyEntry(local_id, entry.name, entry.value,
                                                    NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW, uid)
        return uid
    
    def _getPersistentEntries(self, periodic):
        # copy values out of storage as quickly as possible so lock isn't held
        with self.m_mutex:
            if periodic and not self.m_persistent_dirty:
                return False
            
            self.m_persistent_dirty = False
            
            entries = [(entry.name, entry.value) for entry in self.m_entries.values() \
                       if entry.value is not None and entry.isPersistent()]
        
        # sort in name order
        entries.sort()
        return entries
    
    # ntcore: called getEntries
    def getEntryValues(self, prefix):
        # copy values out of storage as quickly as possible so lock isn't held
        with self.m_mutex:
            entries = [(entry.name, entry.value) for entry in self.m_entries.values() \
                       if entry.value is not None and entry.name.startswith(prefix)]
        
        # sort in name order
        entries.sort()
        return entries
    
    def createRpc(self, local_id, defn, rpc_uid):
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return
            
            old_value = entry.value
            value = Value.makeRpc(defn)
            entry.value = value
        
            # set up the RPC info
            entry.rpc_uid = rpc_uid
            
            if old_value == value:
                return
            
            # assign an id if it doesn't have one
            if entry.id == 0xffff:
                entry.id = len(self.m_idmap)
                self.m_idmap.append(entry)
            
            # generate message
            if outgoing is not None:
                if old_value is None or old_value.type != value.type:
                    entry.increment_seqnum()
                    msg = Message.entryAssign(entry.name, entry.id, entry.seq_num,
                                              value, entry.flags)
                else:
                    entry.increment_seqnum()
                    msg = Message.entryUpdate(entry.id, entry.seq_num, value)
                
                outgoing.append((msg, None, None))
    
    def callRpc(self, local_id, params):
        with self._lockAndGetSendQueue() as outgoing:
            try:
                entry = self.m_localmap[local_id]
            except IndexError:
                return 0
        
            if entry.value is None or not entry.isRpc():
                return 0
        
            entry.rpc_call_uid += 1
            entry.rpc_call_uid &= 0xffff
            call_uid = entry.rpc_call_uid
            
            msg = Message.executeRpc(entry.id, call_uid, params)
            if not self.m_server:
                outgoing.append((msg, None, None))
                return call_uid
            
            # RPCs are unlikely to be used locally on the server, handle it
            # gracefully anyway.
            rpc_uid = entry.rpc_uid
            name = entry.name
        
        conn_info = ConnectionInfo('Server', 'localhost', 0, monotonic(), 0x0300)
        self.m_rpc_server.processRpc(local_id, call_uid, name, msg, conn_info,
                                     self._process_rpc, rpc_uid)
        return call_uid
    
    def _process_rpc(self, local_id, call_uid, result):
        with self.m_mutex:
            self.m_rpc_results[(local_id, call_uid)] = result
            self.m_rpc_results_cond.notify_all()
    
    def getRpcResult(self, local_id, call_uid, timeout=None):
        # returns timed_out, result
        
        with self.m_mutex:
            call_pair = (local_id, call_uid)
            
            # only allow one blocking call per rpc call uid
            if call_pair in self.m_rpc_blocking_calls:
                return False, None
        
            self.m_rpc_blocking_calls.add(call_pair)
            wait_until = monotonic() + timeout
        
            try:
                while True:
                    result = self.m_rpc_results.pop(call_pair, None) 
                    if result is None:
                        if timeout == 0 or self.m_terminating:
                            return False, None
            
                        if timeout < 0:
                            self.m_rpc_results_cond.wait()
                        else:
                            ttw = monotonic() - wait_until
                            if ttw <= 0:
                                return False, None
                            
                            self.m_rpc_results_cond.wait(ttw)
            
                        # if element does not exist, have been canceled
                        if call_pair not in self.m_rpc_blocking_calls:
                            return False, None
            
                        if self.m_terminating:
                            return False, None
            
                        continue
                    
                    return True, result
            finally:
                try:
                    self.m_rpc_blocking_calls.remove(call_pair)
                except KeyError:
                    pass
    
    def cancelBlockingRpcResult(self, call_uid):
        with self.m_mutex:
            # safe to erase even if id does not exist
            self.m_rpc_blocking_calls.erase(call_uid)
            self.m_rpc_results_cond.notify_all()
    
    #
    # Persistence stuff from Storage_load.cpp/Storage_save.cpp
    #
    
    # from Storage_load.cpp
    def loadPersistent(self, filename=None, fp=None):
        return self._loadFromFile(True, '', filename, fp)
        
    def loadEntries(self, filename=None, fp=None, prefix=''):
        return self._loadFromFile(False, prefix, filename, fp)
    
    def _loadFromFile(self, persistent, prefix, filename, fp):
        try:
            if fp:
                entries = load_entries(fp, filename if filename else '<string>', prefix)
            else:
                with open(filename, 'r') as fp:
                    entries = load_entries(fp, filename, prefix)
        except IOError as e:
            return 'Error reading file: %s' % e
        else:
            self._loadEntries(entries, persistent)
            return
    
    def _loadEntries(self, entries, persistent):
        # entries is a list of (str, Value) tuples
        
        # copy values into storage as quickly as possible so lock isn't held
        with self._lockAndGetSendQueue() as outgoing:
            for name, value in entries:
                entry = self._getOrNew(name)
                old_value = entry.value
                entry.value = value
                was_persist = entry.isPersistent()
                if not was_persist and persistent:
                    entry.flags |= NT_PERSISTENT
                
                # if we're the server, an id if it doesn't have one
                if self.m_server and entry.id == 0xffff:
                    entry.id = len(self.m_idmap)
                    self.m_idmap.append(entry)
                
                # notify (for local listeners)
                if self.m_notifier.m_local_notifiers:
                    if old_value is None:
                        self.m_notifier.notifyEntry(entry.local_id, name, value,
                                                    NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
                    elif old_value != value:
                        notify_flags = NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL
                        if not was_persist and persistent:
                            notify_flags |= NT_NOTIFY_FLAGS

                        self.m_notifier.notifyEntry(entry.local_id, name, value, notify_flags)
                    elif not was_persist and persistent:
                        self.m_notifier.notifyEntry(entry.local_id, name, value,
                                                    NT_NOTIFY_FLAGS | NT_NOTIFY_LOCAL)
    
                if outgoing is None:
                    continue    # shortcut
    
                entry.increment_seqnum()
    
                # put on update queue
                if old_value is None or old_value.type != value.type:
                    outgoing.append((Message.entryAssign(name, entry.id,
                                                        entry.seq_num,
                                                        value, entry.flags), None, None))
                elif entry.id != 0xffff:
                    # don't send an update if we don't have an assigned id yet
                    if old_value != value:
                        outgoing.append((Message.entryUpdate(entry.id, entry.seq_num, value), None, None))
                    if not was_persist and persistent:
                        outgoing.append((Message.flagsUpdate(entry.id, entry.flags), None, None))
    
    # from Storage_save.cpp
    def savePersistent(self, filename=None, periodic=False, fp=None):
        entries = self._getPersistentEntries(periodic)
        if entries == False:
            return
        
        err = self._saveEntries(entries, filename, fp)
        if err and periodic:
            self.m_persistent_dirty = True
        return err

    def saveEntries(self, prefix, filename=None, fp=None):
        entries = self.getEntryValues(prefix)
        return self._saveEntries(entries, filename, fp)
        
    def _saveEntries(self, entries, filename, fp):
        # Going to not use tempfile to keep compatibility with ntcore,
        # as having to cleanup temp files on the RIO is probably bad
        if fp:
            try:
                save_entries(fp, entries)
            except IOError as e:
                return 'Error writing file: %s' % e
        else:
            tmp = '%s.tmp' % filename
            bak = '%s.bak' % filename
            
            try:
                with open(tmp, 'w') as fp:
                    save_entries(fp, entries)
                    os.fsync(fp.fileno())
            except IOError as e:
                return 'Error writing file: %s' % e
            
            try:
                file_replace(filename, bak)
            except OSError:
                pass # ignored
                
            try:
                file_replace(tmp, filename)
            except OSError as e:
                try:
                    # try to restore backup
                    file_replace(bak, filename)
                except OSError:
                    pass
                
                return 'Could not rename temp file to real file: %s' % e





# validated: 2017-10-03 DS 1f18cc54167b cpp/ntcore_cpp.cpp

from .connection_notifier import ConnectionNotifier
from .dispatcher import Dispatcher
from .ds_client import DsClient
from .entry_notifier import EntryNotifier
from .rpc_server import RpcServer
from .storage import Storage

from ntcore.constants import NT_NOTIFY_IMMEDIATE, NT_NOTIFY_NEW
_is_new = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW

class NtCoreApi(object):
    '''
        Internal NetworkTables API wrapper
        
        In theory you could create multiple instances of this
        and talk to multiple NT servers or create multiple
        NT servers... though, I don't really know why one
        would want to do this.
    '''
    
    def __init__(self, entry_creator, verbose=False):
        self.conn_notifier = ConnectionNotifier(verbose=verbose)
        self.entry_notifier = EntryNotifier(verbose=verbose)
        self.rpc_server = RpcServer(verbose=verbose)
        self.storage = Storage(self.entry_notifier, self.rpc_server, entry_creator)
        self.dispatcher = Dispatcher(self.storage, self.conn_notifier, verbose=verbose)
        self.ds_client = DsClient(self.dispatcher, verbose=verbose)
        
    def stop(self):
        self.ds_client.stop()
        self.dispatcher.stop()
        self.rpc_server.stop()
        self.entry_notifier.stop()
        self.conn_notifier.stop()
        self.storage.stop()
        
    def destroy(self):
        self.ds_client = None
        self.dispatcher = None
        self.rpc_server = None
        self.entry_notifier = None
        self.entry_notifier = None
        self.conn_notifier = None
        self.storage = None

    #
    # Table functions
    #
    
    def getEntry(self, name):
        return self.storage.getEntry(name)
    
    def getEntryId(self, name):
        return self.storage.getEntryId(name)
    
    def getEntries(self, prefix, types):
        return self.storage.getEntries(prefix, types)
    
    def getEntryNameById(self, local_id):
        return self.storage.getEntryNameById(local_id)
    
    def getEntryTypeById(self, local_id):
        return self.storage.getEntryTypeById(local_id)
    
    def getEntryValue(self, name):
        return self.storage.getEntryValue(name)
    
    def getEntryValueById(self, local_id):
        return self.storage.getEntryValueById(local_id)
    
    def setDefaultEntryValue(self, name, value):
        return self.storage.setDefaultEntryValue(name, value)
    
    def setDefaultEntryValueById(self, local_id, value):
        return self.storage.setDefaultEntryValueById(local_id, value)
    
    def setEntryValue(self, name, value):
        return self.storage.setEntryValue(name, value)
    
    def setEntryValueById(self, local_id, value):
        return self.storage.setEntryValueById(local_id, value)
    
    def setEntryTypeValue(self, name, value):
        self.storage.setEntryTypeValue(name, value)
        
    def setEntryTypeValueById(self, local_id, value):
        self.storage.setEntryTypeValueById(local_id, value)
    
    def setEntryFlags(self, name, flags):
        self.storage.setEntryFlags(name, flags)
        
    def setEntryFlagsById(self, local_id, flags):
        self.storage.setEntryFlagsById(local_id, flags)
    
    def getEntryFlags(self, name):
        return self.storage.getEntryFlags(name)
    
    def getEntryFlagsById(self, local_id):
        return self.storage.getEntryFlags(local_id)
    
    def deleteEntry(self, name):
        self.storage.deleteEntry(name)
    
    def deleteEntryById(self, local_id):
        self.storage.deleteEntry(local_id)
    
    def deleteAllEntries(self):
        self.storage.deleteAllEntries()
    
    def getEntryInfo(self, prefix, types):
        return self.storage.getEntryInfo(prefix, types)
    
    def getEntryInfoById(self, local_id):
        return self.storage.getEntryInfoById(local_id)
    
    #
    # Entry notification
    #
    
    def addEntryListener(self, prefix, callback, flags):
        return self.storage.addListener(prefix, callback, flags)
    
    def addEntryListenerById(self, local_id, callback, flags):
        return self.storage.addListenerById(local_id, callback, flags)
    
    def addEntryListenerByIdEx(self, fromobj, key, local_id, callback, flags, paramIsNew):
        if paramIsNew:
            def listener(item):
                key_, value_, flags_, _ = item
                callback(fromobj, key, value_.value, (flags_ & _is_new) != 0)
        else:
            def listener(item):
                key_, value_, flags_, _ = item
                callback(fromobj, key, value_.value, flags_)
        
        return self.storage.addListenerById(local_id, listener, flags)
    
    def createEntryListenerPoller(self):
        return self.entry_notifier.createPoller()
    
    def destroyEntryListenerPoller(self, poller_uid):
        self.entry_notifier.removePoller(poller_uid)
    
    def addPolledEntryListener(self, poller_uid, prefix, flags):
        return self.storage.addPolledListener(poller_uid, prefix, flags)
    
    def addPolledEntryListenerById(self, poller_uid, local_id, flags):
        return self.storage.addPolledListenerById(poller_uid, local_id, flags)
    
    def pollEntryListener(self, poller_uid, timeout=None):
        return self.entry_notifier.poll(poller_uid, timeout=timeout)
    
    def cancelPollEntryListener(self, poller_uid):
        self.entry_notifier.cancelPoll(poller_uid)
    
    def removeEntryListener(self, listener_uid):
        self.entry_notifier.remove(listener_uid)
    
    def waitForEntryListenerQueue(self, timeout):
        return self.entry_notifier.waitForQueue(timeout)
    
    #
    # Connection notifications
    #
    
    def addConnectionListener(self, callback, immediate_notify):
        return self.dispatcher.addListener(callback, immediate_notify)
    
    def createConnectionListenerPoller(self):
        return self.conn_notifier.createPoller()
    
    def destroyConnectionListenerPoller(self, poller_uid):
        self.conn_notifier.removePoller(poller_uid)
    
    def addPolledConnectionListener(self, poller_uid, immediate_notify):
        return self.dispatcher.addPolledListener(poller_uid, immediate_notify)
    
    def pollConnectionListener(self, poller_uid, timeout=None):
        return self.conn_notifier.poll(poller_uid, timeout=timeout)
    
    def cancelPollConnectionListener(self, poller_uid):
        self.conn_notifier.cancelPoll()
    
    def removeConnectionListener(self, listener_uid):
        self.conn_notifier.remove(listener_uid)
    
    def waitForConnectionListenerQueue(self, timeout):
        return self.conn_notifier.waitForQueue(timeout)

    #
    # TODO: RPC stuff not currently implemented
    #       .. there's probably a good pythonic way to implement
    #          it, but I don't really want to deal with it now.
    #          If you care, submit a PR.
    #
    #          I would have the caller register the server function
    #          via a docstring.
    #
    
    #
    # Client/Server Functions
    #
        
    def setNetworkIdentity(self, name):
        self.dispatcher.setIdentity(name)
    
    def getNetworkMode(self):
        return self.dispatcher.getNetworkMode()
    
    # python-specific
    def startTestMode(self, is_server):
        if self.dispatcher.startTestMode(is_server):
            self.storage.m_server = is_server
            return True
        else:
            return False
        
    def startServer(self, persist_filename, listen_address, port):
        return self.dispatcher.startServer(persist_filename, listen_address, port)
    
    def stopServer(self):
        self.dispatcher.stop()
        
    def startClient(self):
        return self.dispatcher.startClient()
        
    def stopClient(self):
        self.dispatcher.stop()
    
    def setServer(self, server_or_servers):
        self.dispatcher.setServer(server_or_servers)
        
    def setServerTeam(self, teamNumber):
        self.dispatcher.setServerTeam(teamNumber)
    
    def startDSClient(self, port):
        self.ds_client.start(port)
    
    def stopDSClient(self):
        self.ds_client.stop()
    
    def setUpdateRate(self, interval):
        self.dispatcher.setUpdateRate(interval)
    
    def flush(self):
        self.dispatcher.flush()
    
    def getRemoteAddress(self):
        if not self.dispatcher.isServer():
            for conn in self.dispatcher.getConnections():
                return conn.remote_ip
    
    def getIsConnected(self):
        return self.dispatcher.isConnected()
    
    def setVerboseLogging(self, verbose):
        self.conn_notifier.setVerboseLogging(verbose)
        self.dispatcher.setVerboseLogging(verbose)
        self.entry_notifier.setVerboseLogging(verbose)
        self.rpc_server.setVerboseLogging(verbose)
    
    #
    # Persistence
    #
    
    def savePersistent(self, filename):
        return self.storage.savePersistent(filename, periodic=False)
        
    def loadPersistent(self, filename):
        return self.storage.loadPersistent(filename)
    
    def saveEntries(self, filename, prefix):
        return self.storage.saveEntries(prefix, filename=filename)
    
    def loadEntries(self, filename, prefix):
        return self.storage.loadEntries(filename=filename, prefix=prefix)
    
        


# validated: 2017-10-01 DS e4a8bff70e77 cpp/EntryNotifier.cpp cpp/EntryNotifier.h cpp/IEntryNotifier.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import namedtuple

from .callback_manager import CallbackManager, CallbackThread

from .constants import (
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS
)


_EntryListenerData = namedtuple('EntryListenerData', [
    'prefix',
    'local_id', # we don't have entry handles like ntcore has
    'flags',
    'callback',
    'poller_uid',
])
    
# 
_EntryNotification = namedtuple('EntryNotification', [
    'name',
    'value',
    'flags',
    'local_id',
])

_assign_both = NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS
_immediate_local = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_LOCAL


class EntryNotifierThread(CallbackThread):
    
    def __init__(self):
        CallbackThread.__init__(self, 'entry-notifier')
    
    def matches(self, listener, data):
        if not data.value:
            return False
        
        # Flags must be within requested flag set for this listener.
        # Because assign messages can result in both a value and flags update,
        # we handle that case specially.
        listen_flags = listener.flags & ~_immediate_local
        flags = data.flags & ~_immediate_local
  
        if (flags & _assign_both) == _assign_both:
            if (listen_flags & _assign_both) == 0:
                return False
            listen_flags &= ~_assign_both
            flags &= ~_assign_both
        
        if (flags & ~listen_flags) != 0:
            return False

        # must match local id or prefix
        if listener.local_id is not None:
            if listener.local_id != data.local_id:
                return False
        else:
            if not data.name.startswith(listener.prefix):
                return False

        return True
    
    def setListener(self, data, listener_uid):
        pass
        
    def doCallback(self, callback, data):
        callback(data)


class EntryNotifier(CallbackManager):
    
    THREAD_CLASS = EntryNotifierThread
    
    def __init__(self, verbose):
        CallbackManager.__init__(self, verbose)
        
        self.m_local_notifiers = False
    
    def add(self, callback, prefix, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(prefix, None, flags, callback, None))
    
    def addById(self, callback, local_id, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(None, local_id, flags, callback, None))
    
    def addPolled(self, poller_uid, prefix, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(prefix, None, flags, None, poller_uid))
    
    def addPolledById(self, poller_uid, local_id, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(None, local_id, flags, None, poller_uid))
    
    def notifyEntry(self, local_id, name, value, flags, only_listener=None):
        
        # optimization: don't generate needless local queue entries if we have
        # no local listeners (as this is a common case on the server side)
        if not self.m_local_notifiers and (flags & NT_NOTIFY_LOCAL) != 0:
            return
        
        self.send(only_listener, _EntryNotification(name, value, flags, local_id))
    
    def start(self):
        CallbackManager.start(self)



# validated: 2017-10-01 DS 1f18cc54167b cpp/Storage_save.cpp
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import ast
import base64
import re

from .constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
)

from .support.compat import RawConfigParser, PY2

import logging
logger = logging.getLogger('nt')


PERSISTENT_SECTION = 'NetworkTables Storage 3.0'

_key_bool = re.compile('boolean "(.+)"')
_key_double = re.compile('double "(.+)"')
_key_string = re.compile('string "(.+)"')
_key_raw = re.compile('raw "(.+)"')
_key_bool_array = re.compile('array boolean "(.+)"')
_key_double_array = re.compile('array double "(.+)"')
_key_string_array = re.compile('array string "(.+)"')

_value_string = re.compile(r'"((?:\\.|[^"\\])*)",?')

# TODO: these escape functions almost certainly don't deal with unicode
#       correctly

# TODO: strictly speaking, this isn't 100% compatible with ntcore... but

def _unescape_string(s):
    # shortcut if no escapes present
    if '\\' not in s:
        return s
    
    # let python do the hard work
    return ast.literal_eval('"%s"' % s)

# This is mostly what we want... unicode strings won't work properly though
_table = {i: chr(i) if i >= 32 and i < 127 else '\\x%02x' % i for i in range(256)}
_table[ord('"')] = '\\"'
_table[ord('\\')] = '\\\\'
_table[ord('\n')] = '\\n'
_table[ord('\t')] = '\\t'
_table[ord('\r')] = '\\r'

if PY2:
    _table = dict(map(lambda v: (v[0], unicode(v[1])), _table.items()))
    def _escape_string(s):
        return unicode(s).translate(_table)
else:
    def _escape_string(s):
        return s.translate(_table)


def save_entries(fp, entries):
    
    parser = RawConfigParser()
    parser.optionxform = str
    parser.add_section(PERSISTENT_SECTION)
    
    for name, value in entries:
        if not value:
            continue
        
        t = value.type
        v = value.value 
        
        if t == NT_BOOLEAN:
            name = 'boolean "%s"' % _escape_string(name)
            vrepr = 'true' if v else 'false'
        elif t == NT_DOUBLE:
            name = 'double "%s"' % _escape_string(name)
            vrepr = str(v)
        elif t == NT_STRING:
            name = 'string "%s"' % _escape_string(name)
            vrepr = '"%s"' % _escape_string(v)
        elif t == NT_RAW:
            name = 'raw "%s"' % _escape_string(name)
            vrepr = base64.b64encode(v).decode('ascii')
        elif t == NT_BOOLEAN_ARRAY:
            name = 'array boolean "%s"' % _escape_string(name)
            vrepr = ','.join(['true' if vv else 'false' for vv in v])
        elif t == NT_DOUBLE_ARRAY:
            name = 'array double "%s"' % _escape_string(name)
            vrepr = ','.join([str(vv) for vv in v])
        elif t == NT_STRING_ARRAY:
            name = 'array string "%s"' % _escape_string(name)
            vrepr = '","'.join([_escape_string(vv) for vv in v])
            if vrepr:
                vrepr = '"%s"' % vrepr
        else:
            continue
        
        parser.set(PERSISTENT_SECTION, name, vrepr)
    
    if PY2:
        parser.write(fp)
    else:
        parser.write(fp, space_around_delimiters=False)    



# validated: 2017-09-28 DS 5ab20bb27c97 cpp/CallbackManager.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import deque, namedtuple

from .support.compat import Condition, Queue
from .support.safe_thread import SafeThread
from .support.uidvector import UidVector

import logging
logger = logging.getLogger('nt')

_ListenerData = namedtuple('ListenerData',[
    'callback',
    'poller_uid',
])


class Poller(object):
    
    def __init__(self):
        # Note: this is really close to the python queue, but we really have to
        # mess with its internals to get the same semantics as WPILib, so we are
        # rolling our own :(
        self.poll_queue = deque()
        self.poll_cond = Condition()
        self.terminating = False
        self.cancelling = False
    
    def terminate(self):
        with self.poll_cond:
            self.terminating = True
            self.poll_cond.notify_all()


class CallbackThread(object):
    
    def __init__(self, name):
        # Don't need this in python, queue already has one
        #self.m_mutex = threading.Lock()
        
        self.m_listeners = UidVector()
        self.m_queue = Queue()
        self.m_pollers = UidVector()
        
        self.m_active = False
        self.name = name

    #
    # derived must implement the following
    #

    def matches(self, listener, data):
        raise NotImplementedError
    
    def setListener(self, data, listener_uid):
        raise NotImplementedError
    
    def doCallback(self, callback, data):
        raise NotImplementedError

    #
    # Impl
    #
    
    def start(self):
        self.m_active = True
        self._thread = SafeThread(target=self.main, name=self.name)

    def stop(self):
        self.m_active = False
        self.m_queue.put(None)
    
    def sendPoller(self, poller_uid, *args):
        # args are (listener_uid, item)
        poller = self.m_pollers.get(poller_uid)
        if poller:
            with poller.poll_cond:
                poller.poll_queue.append(args)
                poller.poll_cond.notify()
    
    def main(self):
        while self.m_active:
            item = self.m_queue.get()
            if not self.m_active:
                logger.debug('%s thread no longer active', self.name)
                break
            
            if not item:
                continue
            
            listener_uid, item = item
            if listener_uid is not None:
                listener = self.m_listeners.get(listener_uid)
                if listener and self.matches(listener, item):
                    self.setListener(item, listener_uid)
                    if listener.callback:
                        try:
                            self.doCallback(listener.callback, item)
                        except Exception:
                            logger.warn("Unhandled exception processing %s callback", self.name, exc_info=True)
                    elif listener.poller_uid is not None:
                        self.sendPoller(listener.poller_uid, listener_uid, item)
            else:
                # Use copy because iterator might get invalidated
                for listener_uid, listener in list(self.m_listeners.items()):
                    if self.matches(listener, item):
                        self.setListener(item, listener_uid)
                        if listener.callback:
                            try:
                                self.doCallback(listener.callback, item)
                            except Exception:
                                logger.warn("Unhandled exception processing %s callback", self.name, exc_info=True)
                        elif listener.poller_uid is not None:
                            self.sendPoller(listener.poller_uid, listener_uid, item)
        
        # Wake any blocked pollers
        for poller in self.m_pollers.values():
            poller.terminate()

class CallbackManager(object):
    
    # Derived classes should declare this attribute at class level:
    # THREAD_CLASS = Something
    
    def __init__(self, verbose):
        self.m_verbose = verbose
        self.m_owner = None
    
    def setVerboseLogging(self, verbose):
        self.m_verbose = verbose
    
    def stop(self):
        if self.m_owner:
            self.m_owner.stop()
    
    def remove(self, listener_uid):
        thr = self.m_owner
        if thr:
            thr.m_listeners.pop(listener_uid, None)
    
    def createPoller(self):
        self.start()
        thr = self.m_owner
        return thr.m_pollers.add(Poller())
    
    def removePoller(self, poller_uid):
        thr = self.m_owner
        if not thr:
            return
        
        # Remove any listeners that are associated with this poller
        listeners = list(thr.m_listeners.items())
        for lid, listener in listeners:
            if listener.poller_uid == poller_uid:
                thr.m_listeners.pop(lid)
        
        # Wake up any blocked pollers
        poller = thr.m_pollers.pop(poller_uid, None)
        if not poller:
            return
        
        poller.terminate()
    
    def waitForQueue(self, timeout):
        thr = self.m_owner
        if not thr:
            return True
        
        q = thr.m_queue
        
        def _poll_fn():
            if q._qsize() == 0:
                return True
            if not thr.m_active:
                return True
            return False
                    
        with q.not_full:
            return q.not_full.wait_for(_poll_fn, timeout)
    
    def poll(self, poller_uid, timeout=None):
        # returns infos, timed_out
        # -> infos is a list of (listener_uid, item)
        infos = []
        timed_out = False
        
        thr = self.m_owner
        if not thr:
            return infos, timed_out
        
        poller = thr.m_pollers.get(poller_uid)
        if not poller:
            return infos, timed_out
        
        def _poll_fn():
            if poller.poll_queue:
                return 1
            if poller.cancelling:
                # Note: this only works if there's a single thread calling this
                # function for any particular poller, but that's the intended use.
                poller.cancelling = False
                return 2
        
        with poller.poll_cond:
            result = poller.poll_cond.wait_for(_poll_fn, timeout)
            if result is None: # timeout
                timed_out = True
            elif result == 1: # success
                infos.extend(poller.poll_queue)
                poller.poll_queue.clear()
        
        return infos, timed_out
    
    def cancelPoll(self, poller_uid):
        thr = self.m_owner.getThread()
        if not thr:
            return
            
        poller = thr.m_pollers.get(poller_uid)
        if not poller:
            return
        
        with poller.poll_cond:
            poller.cancelling = True
            poller.poll_cond.notify()
    
    # doStart in ntcore
    def start(self, *args):
        if not self.m_owner:
            self.m_owner = self.THREAD_CLASS(*args)
            self.m_owner.start()
    
    # Unlike ntcore, only a single argument is supported here. This is 
    # to ensure that it's more clear what struct is being passed through
    
    def doAdd(self, item):
        self.start()
        thr = self.m_owner
        return thr.m_listeners.add(item)
        
    def send(self, only_listener, item):
        thr = self.m_owner
        if not thr or not thr.m_listeners:
            return
        
        thr.m_queue.put((only_listener, item))



# validated: 2017-10-01 DS e68a71022c96 cpp/ConnectionNotifier.cpp cpp/ConnectionNotifier.h cpp/IConnectionNotifier.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import namedtuple

from .callback_manager import CallbackManager, CallbackThread

_ConnectionCallback = namedtuple('ConnectionCallback', [
    'callback',
    'poller_uid',
])

_ConnectionNotification = namedtuple('ConnectionNotification', [
    'connected',
    'conn_info',
])


class ConnectionNotifierThread(CallbackThread):
    
    def __init__(self):
        CallbackThread.__init__(self, 'connection-notifier')
    
    def matches(self, listener, data):
        return True
    
    def setListener(self, data, listener_uid):
        pass
        
    def doCallback(self, callback, data):
        callback(data)


class ConnectionNotifier(CallbackManager):
    
    THREAD_CLASS = ConnectionNotifierThread
    
    def add(self, callback):
        return self.doAdd(_ConnectionCallback(callback, None))
        
    def addPolled(self, poller_uid):
        return self.doAdd(_ConnectionCallback(None, poller_uid))
    
    def notifyConnection(self, connected, conn_info, only_listener=None):
        self.send(only_listener, _ConnectionNotification(connected, conn_info))

    def start(self):
        CallbackManager.start(self)




import threading

import logging
logger = logging.getLogger('nt.th')


class SafeThread(object):
    '''
        Not exactly the same as wpiutil SafeThread... exists so we don't have
        to duplicate functionality in a lot of places
    '''
    
    # Name each thread uniquely to make debugging easier
    _global_indices_lock = threading.Lock()
    _global_indices = {}
    
    def __init__(self, target, name, args=()):
        '''
            Note: thread is automatically started and daemonized
        '''
        
        with SafeThread._global_indices_lock:
            idx = SafeThread._global_indices.setdefault(name, -1) + 1
            SafeThread._global_indices[name] = idx
            name = '%s-%s' % (name, idx)
        
        self.name = name
        
        self._thread = threading.Thread(target=self._run, name=name, args=(target, args))
        self._thread.daemon = True
        
        self.is_alive = self._thread.is_alive
        self.join = self._thread.join
            
        self._thread.start()
    
    def join(self, timeout=1):
        self._thread.join(timeout=timeout)
        if not self._thread.is_alive():
            logger.warn('Thread %s did not stop!', self.name)
    
    def _run(self, target, args):
        logger.debug("Started thread %s", self.name)
        try:
            target(*args)
        except Exception:
            logger.warn("Thread %s died unexpectedly", self.name, exc_info=True)
        else:
            logger.debug("Thread %s exited", self.name)



# novalidate

import threading

class UidVector(dict):
    
    def __init__(self):
        self.idx = 0
        self.lock = threading.Lock()
    
    def add(self, item):
        '''Only use this method to add to the UidVector'''
        with self.lock:
            idx = self.idx
            self.idx += 1
        
        self[idx] = item
        return idx



# validated: 2017-10-01 DS 5ab20bb27c97 cpp/RpcServer.cpp cpp/RpcServer.h cpp/IRpcServer.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

from collections import namedtuple

from .callback_manager import CallbackManager, CallbackThread
from .message import Message

import logging
logger = logging.getLogger('nt')

_RpcListenerData = namedtuple('RpcListenerData', [
    'callback',
    'poller_uid',
])

_RpcCall = namedtuple('RpcCall', [
    'local_id',
    'call_uid',
    'name',
    'params',
    'conn_info',
    'send_response',
])


class RpcServerThread(CallbackThread):
    
    def __init__(self, inst):
        CallbackThread.__init__(self, 'rpc-server')
        self.m_inst = inst
        self.m_response_map = {}
    
    def matches(self, listener, data):
        return data.name and data.send_response
    
    def setListener(self, data, listener_uid):
        lookup_id = (data.local_id, data.call_uid)
        self.m_response_map[lookup_id] = data.send_response
        
    def doCallback(self, callback, data):
        local_id = data.local_id
        call_uid = data.call_uid
        lookup_id = (data.local_id, data.call_uid)
        callback(data)
        
        # send empty response
        send_response = self.m_response_map.get(lookup_id)
        if send_response:
            send_response(Message.rpcResponse(local_id, call_uid, ""))


class RpcServer(CallbackManager):
    
    THREAD_CLASS = RpcServerThread
    
    def add(self, callback):
        return self.doAdd(_RpcListenerData(callback, None))
    
    def addPolled(self, poller_uid):
        return self.doAdd(_RpcListenerData(None, poller_uid))
    
    def removeRpc(self, rpc_uid):
        return self.remove(rpc_uid)
    
    def processRpc(self, local_id, call_uid, name, params, conn_info, send_response, rpc_uid):
        call = _RpcCall(local_id, call_uid, name, params, conn_info, send_response)
        self.send(rpc_uid, call)
    
    def postRpcResponse(self, local_id, call_uid, result):
        thr = self.m_owner
        response = thr.m_response_map.pop((local_id, call_uid), None)
        if response is None:
            logger.warning("Posting RPC response to nonexistent call (or duplicate response)")
        else:
            response(Message.rpcResponse(local_id, call_uid, result))
    
    def start(self):
        CallbackManager.start(self.m_inst)



# validated: 0000-00-00 DS 000000000 cpp/DsClient.cpp cpp/DsClient.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2017. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import logging
logger = logging.getLogger('nt')

class DsClient(object):
    
    def __init__(self, dispatcher, verbose=False):
        pass
    
    def start(self, port):
        logging.warn("Dashboard client mode not currently implemented")
    
    def stop(self):
        pass




# This is imported first to avoid circular dependency problems
from .version import __version__

#: Alias of NetworkTablesInstance.getDefault()
from .networktables import NetworkTables

from .instance import NetworkTablesInstance




# validated: 2017-10-08 DS 8a37b81f4e10 cpp/Storage_load.cpp

import ast
import binascii
import base64
import re

from .value import Value

from .support.compat import RawConfigParser, NoSectionError, PY2

import logging
logger = logging.getLogger('nt')


PERSISTENT_SECTION = 'NetworkTables Storage 3.0'

_key_bool = re.compile('boolean "(.+)"')
_key_double = re.compile('double "(.+)"')
_key_string = re.compile('string "(.+)"')
_key_raw = re.compile('raw "(.+)"')
_key_bool_array = re.compile('array boolean "(.+)"')
_key_double_array = re.compile('array double "(.+)"')
_key_string_array = re.compile('array string "(.+)"')

_value_string = re.compile(r'"((?:\\.|[^"\\])*)",?')

# TODO: these escape functions almost certainly don't deal with unicode
#       correctly

# TODO: strictly speaking, this isn't 100% compatible with ntcore... but

def _unescape_string(s):
    # shortcut if no escapes present
    if '\\' not in s:
        return s
    
    # let python do the hard work
    return ast.literal_eval('"%s"' % s)


def load_entries(fp, filename, prefix):
    
    entries = []
    
    parser = RawConfigParser()
    parser.optionxform = str
    
    try:
        if hasattr(parser, 'read_file'):
            parser.read_file(fp, filename)
        else:
            parser.readfp(fp, filename)
    except IOError:
        raise
    except Exception as e:
        raise IOError('Error reading persistent file: %s' % e)
    
    try:
        items = parser.items(PERSISTENT_SECTION)
    except NoSectionError:
        raise IOError("Persistent section not found")

    value = None
    m = None
    
    for k, v in items:
        
        # Reduces code duplication
        if value:
            key = _unescape_string(m.group(1))
            if key.startswith(prefix):
                entries.append((key, value))
            
        value = None
        
        m = _key_bool.match(k)
        if m:
            if v == 'true':
                value = Value.makeBoolean(True)
            elif v == 'false':
                value = Value.makeBoolean(False)
            else:
                logger.warn("Unrecognized boolean value %r for %s", v, m.group(1))
            continue
        
        m = _key_double.match(k)
        if m:
            try:
                value = Value.makeDouble(float(v))
            except ValueError as e:
                logger.warn("Unrecognized double value %r for %s", v, m.group(1))
                
            continue
        
        m = _key_string.match(k)
        if m:
            mm = _value_string.match(v)
            
            if mm:
                value = Value.makeString(_unescape_string(mm.group(1)))
            elif PY2 and v == '':
                value = Value.makeString('')
            else:
                logger.warn("Unrecognized string value %r for %s", v, m.group(1))
            continue
        
        m = _key_raw.match(k)
        if m:
            try:
                if PY2:
                    v = base64.b64decode(v)
                else:
                    v = base64.b64decode(v, validate=True)
                value = Value.makeRaw(v)
            except binascii.Error:
                logger.warn("Unrecognized raw value %r for %s", v, m.group(1))
            continue
        
        m = _key_bool_array.match(k)
        if m:
            bools = []
            arr = v.strip().split(',')
            if arr != ['']:
                for vv in arr:
                    vv = vv.strip()
                    if vv == 'true':
                        bools.append(True)
                    elif vv == 'false':
                        bools.append(False)
                    else:
                        logger.warn("Unrecognized bool '%s' in bool array %s'", vv, m.group(1))
                        bools = None
                        break
                
            if bools is not None:
                value = Value.makeBooleanArray(bools)    
            continue
        
        m = _key_double_array.match(k)
        if m:
            doubles = []
            arr = v.strip().split(',')
            if arr != ['']:
                for vv in arr:
                    try:
                        doubles.append(float(vv))
                    except ValueError:
                        logger.warn("Unrecognized double '%s' in double array %s", vv, m.group(1))
                        doubles = None
                        break
                
            value = Value.makeDoubleArray(doubles)    
            continue
        
        m = _key_string_array.match(k)
        if m:
            # Technically, this will let invalid inputs in... but,
            # I don't really care. Feel free to fix it if you do.
            strings = [_unescape_string(vv) for vv in _value_string.findall(v)]
            value = Value.makeStringArray(strings)
            continue
        
        logger.warn("Unrecognized type '%s'", k)
        
    if value:
        key = _unescape_string(m.group(1))
        if key.startswith(prefix):
            entries.append((key, value))
    
    return entries



#!/usr/bin/env python3
#
# Simple benchmark script that tests NetworkTables throughput.
#
# To run the server
#
#     python3 benchmark.py
# 
# To run the client:
#
#    python3 benchmark.py 127.0.0.1
#
# In theory, the limiting factor is internal buffering and not network
# bandwidth, so running this test on a local machine should be mostly
# valid.
#
# On OSX with default write flush period, I average ~18hz
#

from __future__ import print_function

from argparse import ArgumentParser
import time

import logging
logging.basicConfig(level=logging.DEBUG)

from networktables import NetworkTables

class Benchmark(object):

    def __init__(self):
        
        parser = ArgumentParser()
        parser.add_argument('client', nargs='?', default=None)
        parser.add_argument('-r', '--rate', default=None, type=float)
        parser.add_argument('--send', default=False, action='store_true')
        
        args = parser.parse_args()
        
        if args.client:
            NetworkTables.initialize(server=args.client)
        else:
            NetworkTables.initialize()
         
        # Default write flush is 0.05, could adjust for less latency
        if args.rate is not None:
            print("Setting rate to %s" % args.rate)
            NetworkTables.setUpdateRate(args.rate)
        
        self.nt = NetworkTables.getTable('/benchmark')
        self.updates = 0
        
        self.nt.addTableListener(self.on_update)
        
        if args.send:
            self.send_benchmark()
        else:
            self.recv_benchmark()

    def on_update(self, *args):
        self.updates += 1

    def recv_benchmark(self):
        
        print("Starting to receive")
        
        last = None
        last_updates = 0
        
        while True:
            
            now = time.time()
            updates = self.updates
            
            if last is not None:
                rate = (updates - last_updates) / (now - last)
                print('Update rate:', rate)
            
            last = now
            last_updates = updates
            time.sleep(1)
            
    def send_benchmark(self):
        
        print("Sending")
        
        i = 0
        
        while True:
            i += 1
            self.nt.putNumber('key', i)
            time.sleep(0.0001)
    
if __name__ == '__main__':
    Benchmark()